// listplug.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "comdef.h"
#include "afxtempl.h"
#include "afxrich.h"
#include "ListerPlugin.h"
#include "listplug.h"

CConverter* CConverter::m_pThis = NULL;

#define BUFFSIZE 4096

HGLOBAL CConverter::StringToHGLOBAL(LPCSTR pstr)
{
	HGLOBAL hMem = NULL;
	if (pstr != NULL)
	{
		hMem = GlobalAlloc(GHND, (lstrlenA(pstr)*2)+1);
		char* p = (char*) GlobalLock(hMem);
		ASSERT(p != NULL);
		if (p != NULL)
			lstrcpyA(p, pstr);
		GlobalUnlock(hMem);
	}
	return hMem;
}

CConverter::CConverter(LPCSTR pszLibName, HWND hWnd)
{
	USES_CONVERSION;
	m_hBuff = NULL;
	m_pBuf = NULL;
	m_nPercent = 0;
	m_bDone = TRUE;
	m_bConvErr = FALSE;
	m_hFileName = NULL;
	m_hWnd = hWnd;
	OFSTRUCT ofs;
	if (OpenFile(pszLibName, &ofs, OF_EXIST) == HFILE_ERROR)
	{
		m_hLibCnv = NULL;
		return;
	}
	m_hLibCnv = LoadLibraryA(pszLibName);
	if (m_hLibCnv < (HINSTANCE)HINSTANCE_ERROR)
		m_hLibCnv = NULL;
	else
	{
		LoadFunctions();
		ASSERT(m_pInitConverter != NULL);
		if (m_pInitConverter != NULL)
		{
			CString str = "LISTER";
			VERIFY(m_pInitConverter(m_hWnd, T2CA(str)));
		}
	}
}

CConverter::CConverter(HWND hWnd)
{
	m_hWnd = hWnd;
	m_pInitConverter = NULL;
	m_pIsFormatCorrect = NULL;
	m_pForeignToRtf = NULL;
	m_pRtfToForeign = NULL;
	m_bConvErr = FALSE;
	m_hFileName = NULL;
}

CConverter::~CConverter()
{
	Cleanup();
}

void CConverter::Cleanup()
{
	FreeLibrary(m_hLibCnv);
}

BOOL CConverter::IsFormatCorrect(LPCTSTR pszFileName)
{
	USES_CONVERSION;
	int nRet;
	if (m_hLibCnv == NULL || m_pIsFormatCorrect == NULL)
		return FALSE;

	char buf[_MAX_PATH];
	strcpy(buf, T2CA(pszFileName));

	CharToOemA(buf, buf);

	HGLOBAL hFileName = StringToHGLOBAL(buf);
	HGLOBAL hDesc = GlobalAlloc(GHND, 256);
	ASSERT(hDesc != NULL);
	nRet = m_pIsFormatCorrect(hFileName, hDesc);
	GlobalFree(hDesc);
	GlobalFree(hFileName);
	return (nRet == 1) ? TRUE : FALSE;
}

// static callback function
int CALLBACK CConverter::WriteOutStatic(int cch, int nPercentComplete)
{
	ASSERT(m_pThis != NULL);
	return m_pThis->WriteOut(cch, nPercentComplete);
}

int CALLBACK CConverter::WriteOut(int cch, int nPercentComplete)
{
	ASSERT(m_hBuff != NULL);
	m_nPercent = nPercentComplete;
	if ( m_hBuff == NULL ) {
		return -9;
	}
	if (cch != 0)
	{
		int err = 0;
		try {
			Write(m_pBuf, cch);
			err = GetLastError();
		}
		catch (...) {
		}
		return cch;
	}
	return 0; //everything OK
}

BOOL CConverter::DoConversion()
{
	USES_CONVERSION;
	m_nPercent = 0;

	ASSERT(m_hBuff != NULL);
	ASSERT(m_pThis != NULL);
	HGLOBAL hDesc = StringToHGLOBAL("");
	HGLOBAL hSubset = StringToHGLOBAL("");

	ASSERT(m_pForeignToRtf != NULL);
	ASSERT(m_hFileName != NULL);
	int nRet = m_pForeignToRtf(m_hFileName, NULL, m_hBuff, hDesc, hSubset, (LPFNOUT)WriteOutStatic);

	GlobalFree(hDesc);
	GlobalFree(hSubset);
	if (m_pBuf != NULL) {
		GlobalUnlock(m_hBuff);
	}
	GlobalFree(m_hBuff);

	if ( nRet != 0 ) {
		m_bConvErr = TRUE;
	}

	m_bDone = TRUE;
	m_nPercent = 100;

	return (nRet == 0);
}

void CConverter::LoadFunctions()
{
	m_pInitConverter = (PINITCONVERTER)GetProcAddress(m_hLibCnv, "InitConverter32");
	m_pIsFormatCorrect = (PISFORMATCORRECT)GetProcAddress(m_hLibCnv, "IsFormatCorrect32");
	m_pForeignToRtf = (PFOREIGNTORTF)GetProcAddress(m_hLibCnv, "ForeignToRtf32");
	m_pRtfToForeign = (PRTFTOFOREIGN)GetProcAddress(m_hLibCnv, "RtfToForeign32");
}

///////////////////////////////////////////////////////////////////////////////

BOOL CConverter::Open(LPCTSTR pszFileName, UINT nOpenFlags,
	CFileException* pException)
{
	USES_CONVERSION;
	// we convert to oem and back because of the following case
	// test(c).txt becomes testc.txt in OEM and stays testc.txt to Ansi
	char buf[_MAX_PATH];
	strcpy(buf, T2CA(pszFileName));
	CharToOemA(buf, buf);
	OemToCharA(buf, buf);

	LPTSTR lpszFileNameT = A2T(buf);

	BOOL bOpen = CFile::Open(lpszFileNameT, CFile::modeRead, pException);
	CFile::Close();
	if ( !bOpen ) {
		return FALSE;
	}

	char temp[MAX_PATH], tempfn[MAX_PATH];
	GetTempPath(MAX_PATH, temp);
	GetTempFileName(temp, "rtf", 0, tempfn);

	CharToOemA(buf, buf);
	ASSERT(m_hFileName == NULL);
	m_hFileName = StringToHGLOBAL(buf);

	m_pThis = this;
	m_hBuff = GlobalAlloc(GHND, BUFFSIZE);
	ASSERT(m_hBuff != NULL);
	m_pBuf = (unsigned char *)GlobalLock(m_hBuff);

	try {
		CFile::Open(tempfn, CFile::modeReadWrite|CFile::shareExclusive, pException);
		DoConversion();
		CFile::Close();
		CFile::Open(tempfn, CFile::modeRead|CFile::shareExclusive, pException);
	}
	catch (...) {
		CFile::Close();
		return FALSE;
	}

	return TRUE;
}

void CConverter::Close()
{
	CString name = GetFilePath();
	CFile::Close();
	Remove(name);
}

struct ConvType
{
	CString m_ext, m_path, m_signature;
	CConverter * m_conv;

	ConvType(CString ext, CString path, CString signature)
	{
		m_signature = signature;
		m_ext = ext;
		m_path = path;
		m_conv = NULL;
	}

	void ClearConverter()
	{
		m_conv = NULL;
	}

	CConverter * GetConverter()
	{
		if ( !m_conv ) {
			m_conv = new CConverter(m_path);
			if ( !m_conv->m_hLibCnv ) {
				delete m_conv;
				m_conv = NULL;
			}
		}
		return m_conv;
	}
};

CList<ConvType *, ConvType *> ConvMap;

/////////////////////////////////////////////////////////////////////////////

static BOOL IsLeadMatch(CFile& file, const BYTE* pb, UINT nCount)
{
	// check for match at beginning of file
	BOOL b = FALSE;
	BYTE* buf = new BYTE[nCount];

	TRY
	{
		file.SeekToBegin();
		memset(buf, 0, nCount);
		file.Read(buf, nCount);
		if (memcmp(buf, pb, nCount) == 0)
			b = TRUE;
	}
	END_TRY

	delete [] buf;
	return b;
}

CFile * GetDocTypeFromName(HWND hWnd, LPCTSTR pszPathName, CFileException& fe)
{
	CFile file;
	ASSERT(pszPathName != NULL);

	if (!file.Open(pszPathName, CFile::modeRead | CFile::shareDenyWrite, &fe))
		return NULL;
	file.Close();

	char ext[MAX_PATH];
	_splitpath(pszPathName, NULL, NULL, NULL, ext);

	if ( stricmp(ext, ".rtf" ) == 0 ) {
		return new CFile;
	}

	for ( POSITION pos = ConvMap.GetHeadPosition(); pos; ) {
		ConvType * pType = ConvMap.GetNext(pos);
		if ( pType->m_ext.CompareNoCase(ext + 1) ) {
			continue;
		}
		if ( pType->m_signature != "" ) {
			char tmp[80];
			int pos = 1, n = pType->m_signature.GetLength() - 1;
			if ( pType->m_signature[0] == '\"' ) {
				if ( pType->m_signature[n] == '\"' ) {
					--n;
				}
				strncpy(tmp, pType->m_signature.Mid(pos, n), sizeof(tmp));
			} else {
				n = 0;
				for ( const char * s = pType->m_signature; *s; s += 2 ) {
					sscanf(s, "%2x", tmp + n++);
				}
			}
			file.Open(pszPathName, CFile::modeRead | CFile::shareDenyWrite, &fe);
			BOOL bMatch = IsLeadMatch(file, (unsigned char *)tmp, n);
			file.Close();
			if ( !bMatch ) {
				continue;
			}
		}	
		CConverter * pConv = pType->GetConverter();
		if ( pConv && pConv->IsFormatCorrect(pszPathName) ) {
			pType->ClearConverter();
			return pConv;
		}
	}
	return NULL;
}

void ScanForConverters()
{
	static BOOL bInit = FALSE;
	if ( bInit ) {
		return;
	}
	bInit = TRUE;

	char name[20], conv[2*MAX_PATH+2], fname[MAX_PATH], drive[MAX_PATH], dir[MAX_PATH], fn[MAX_PATH], ext[MAX_PATH];
	extern HINSTANCE hinst;
	GetModuleFileName(hinst, fname, sizeof(fname));
	_splitpath(fname, drive, dir, fn, ext);
	_makepath(fname, drive, dir, "office", ".ini");
	for ( int i = 0; ; ++i ) {
		sprintf(name, "%d", i);
		GetPrivateProfileString("converters", name, "", conv, sizeof(conv), fname);
		if ( !*conv ) {
			break;
		}
		char * signature = strchr(conv, ',');
		if ( !signature ) {
			break;
		}
		*signature++ = 0;
		char *path = strchr(signature, ',');
		if ( !path ) {
			break;
		}
		*path++ = 0;
		_splitpath(path, drive, dir, fn, ext);
		BOOL bValid = TRUE;
		if ( !*drive && !*dir ) {
			bValid = FALSE;
			char fname[_MAX_PATH];
			GetModuleFileName(hinst, fname, sizeof(fname));
			_splitpath(fname, drive, dir, NULL, NULL);
			_makepath(path, drive, dir, fn, ext);
			OFSTRUCT ofs;
			if (OpenFile(path, &ofs, OF_EXIST) == HFILE_ERROR) {
				HKEY hKey, hSubKey;
				LONG rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Shared Tools\\Text Converters\\Import", 0, KEY_READ, &hKey);
				if ( rc != ERROR_SUCCESS ) {
					continue;
				}
				char subkey[_MAX_PATH];
				FILETIME ft;
				for (DWORD i = 0; !bValid; ++i ) {
					DWORD size = sizeof(subkey);
					if ( ERROR_SUCCESS != RegEnumKeyEx(hKey, i, subkey, &size, FALSE, NULL, 0, &ft) ) {
						break;
					}
					LONG rc = RegOpenKeyEx(hKey, subkey, 0, KEY_READ, &hSubKey);
					if ( rc != ERROR_SUCCESS ) {
						continue;
					}
					size = sizeof(fname);
					rc = RegQueryValueEx(hSubKey, "Path", NULL, NULL, (unsigned char *)fname, &size);
					if ( rc == ERROR_SUCCESS ) {
						_splitpath(fname, NULL, NULL, drive, dir);
						if ( stricmp(drive, fn) == 0 && stricmp(dir, ext) == 0 ) {
							_splitpath(fname, drive, dir, NULL, NULL);
							bValid = TRUE;
						}
					}
					RegCloseKey(hSubKey);
				}
				RegCloseKey(hKey);
			}
		}
		if ( bValid ) {
			_makepath(path, drive, dir, fn, ext);
			ConvType * pType = new ConvType(conv, path, signature);
			ConvMap.AddTail(pType);
		}
	}
}

HINSTANCE hinst;
HMODULE FLibHandle = 0;

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	switch ( ul_reason_for_call )
	{
		case DLL_PROCESS_ATTACH:
			hinst = (HINSTANCE)hModule;
			OleInitialize(NULL);
			break;

		case DLL_PROCESS_DETACH:
			if ( FLibHandle ) {
				FreeLibrary(FLibHandle);
				OleUninitialize();
			}
			FLibHandle = NULL;
			break;

		case DLL_THREAD_ATTACH:
			break;

		case DLL_THREAD_DETACH:
			break;
	}
	return TRUE;
}

struct OleCallback : public IRichEditOleCallback
{
	int m_nRefCount;

	OleCallback()
	{
		m_nRefCount = 0;
	}

	// *** IUnknown methods ***
	STDMETHOD(QueryInterface) (THIS_ REFIID riid, LPVOID FAR * lplpObj);
	STDMETHOD_(ULONG,AddRef) (THIS);
	STDMETHOD_(ULONG,Release) (THIS);

	// *** IRichEditOleCallback methods ***
	STDMETHOD(GetNewStorage) (THIS_ LPSTORAGE FAR * lplpstg);
	STDMETHOD(GetInPlaceContext) (THIS_ LPOLEINPLACEFRAME FAR * lplpFrame,
							  LPOLEINPLACEUIWINDOW FAR * lplpDoc,
							  LPOLEINPLACEFRAMEINFO lpFrameInfo);
	STDMETHOD(ShowContainerUI) (THIS_ BOOL fShow);
	STDMETHOD(QueryInsertObject) (THIS_ LPCLSID lpclsid, LPSTORAGE lpstg,
								LONG cp);
	STDMETHOD(DeleteObject) (THIS_ LPOLEOBJECT lpoleobj);
	STDMETHOD(QueryAcceptData) (THIS_ LPDATAOBJECT lpdataobj,
							CLIPFORMAT FAR * lpcfFormat, DWORD reco,
							BOOL fReally, HGLOBAL hMetaPict);
	STDMETHOD(ContextSensitiveHelp) (THIS_ BOOL fEnterMode);
	STDMETHOD(GetClipboardData) (THIS_ CHARRANGE FAR * lpchrg, DWORD reco,
								LPDATAOBJECT FAR * lplpdataobj);
	STDMETHOD(GetDragDropEffect) (THIS_ BOOL fDrag, DWORD grfKeyState,
								LPDWORD pdwEffect);
	STDMETHOD(GetContextMenu) (THIS_ WORD seltype, LPOLEOBJECT lpoleobj,
								CHARRANGE FAR * lpchrg,
								HMENU FAR * lphmenu);
};

STDMETHODIMP_(ULONG) OleCallback::AddRef()
{
	return (ULONG)++m_nRefCount;
}

STDMETHODIMP_(ULONG) OleCallback::Release()
{
	if ( --m_nRefCount == 0 ) {
		delete this;
	}
	return (ULONG)m_nRefCount;
}

STDMETHODIMP OleCallback::QueryInterface(
	REFIID iid, LPVOID* ppvObj)
{
	if ( iid == __uuidof(IRichEditOleCallback) ) {
		*ppvObj = (LPVOID)this;
		return S_OK;
	}
	return E_NOINTERFACE;
}

STDMETHODIMP OleCallback::GetNewStorage(LPSTORAGE* ppstg)
{
	// Create a flat storage and steal it from the client item
	// the client item is only used for creating the storage
	COleClientItem item;
	item.GetItemStorageFlat();
	*ppstg = item.m_lpStorage;
	HRESULT hRes = E_OUTOFMEMORY;
	if (item.m_lpStorage != NULL)
	{
		item.m_lpStorage = NULL;
		hRes = S_OK;
	}
	return hRes;
}

STDMETHODIMP OleCallback::GetInPlaceContext(LPOLEINPLACEFRAME* lplpFrame, LPOLEINPLACEUIWINDOW* lplpDoc,
															LPOLEINPLACEFRAMEINFO lpFrameInfo)
{
	return E_NOTIMPL;
}

STDMETHODIMP OleCallback::ShowContainerUI(BOOL fShow)
{
	return E_NOTIMPL;
}

STDMETHODIMP OleCallback::QueryInsertObject(LPCLSID /*lpclsid*/, LPSTORAGE /*pstg*/, LONG /*cp*/)
{
	return S_OK;
}

STDMETHODIMP OleCallback::DeleteObject(LPOLEOBJECT /*lpoleobj*/)
{
	return S_OK;
}

STDMETHODIMP OleCallback::QueryAcceptData(LPDATAOBJECT lpdataobj, CLIPFORMAT* lpcfFormat, DWORD reco, BOOL fReally, HGLOBAL hMetaPict)
{
	return E_NOTIMPL;
}

STDMETHODIMP OleCallback::ContextSensitiveHelp(BOOL /*fEnterMode*/)
{
	return E_NOTIMPL;
}

STDMETHODIMP OleCallback::GetClipboardData(
	CHARRANGE* lpchrg, DWORD reco, LPDATAOBJECT* lplpdataobj)
{
	return E_NOTIMPL;
}

STDMETHODIMP OleCallback::GetDragDropEffect(
	BOOL fDrag, DWORD grfKeyState, LPDWORD pdwEffect)
{
	return E_NOTIMPL;
}

STDMETHODIMP OleCallback::GetContextMenu(
	WORD seltype, LPOLEOBJECT lpoleobj, CHARRANGE* lpchrg,
	HMENU* lphmenu)
{
	return E_NOTIMPL;
}

// return 0 for no error, otherwise return error code
DWORD CALLBACK EditStreamCallBack(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb)
{
	CFile * pFile = (CFile *)dwCookie;
	*pcb = pFile->ReadHuge(pbBuff, cb);
	return *pcb == 0;
}

// return child window handle, if recognize file type, or NULL
HWND __stdcall ListLoad(HWND ParentWin,char* FileToLoad,int ShowFlags)
{
	HWND hwnd;
	RECT r;

	try {
		if (!FLibHandle) {
			int OldError = SetErrorMode(SEM_NOOPENFILEERRORBOX);
			FLibHandle = LoadLibrary("Riched20.dll");
			if ((int)FLibHandle < HINSTANCE_ERROR) 
				FLibHandle = LoadLibrary("RICHED32.DLL");
			if ((int)FLibHandle < HINSTANCE_ERROR) 
				FLibHandle = NULL;
			SetErrorMode(OldError);
		}

		GetClientRect(ParentWin,&r);
		hwnd = CreateWindow("RichEdit20A","",WS_CHILD | WS_VISIBLE | ES_MULTILINE | ES_READONLY | WS_VSCROLL | ES_NOHIDESEL,
																	r.left,r.top,r.right-r.left, r.bottom-r.top,ParentWin,NULL,hinst,NULL);
		
		if ( !hwnd ) {
			hwnd = CreateWindow("RichEdit","",WS_CHILD | WS_VISIBLE | ES_MULTILINE | ES_READONLY | WS_VSCROLL | ES_NOHIDESEL,
																	r.left,r.top,r.right-r.left, r.bottom-r.top,ParentWin,NULL,hinst,NULL);
		}

		if ( !IsWindow(hwnd) ) {
			return NULL;
		}

		ScanForConverters();

		CFileException fe;
		CFile * pFile = GetDocTypeFromName(ParentWin, FileToLoad, fe);
		if ( !pFile || !pFile->Open(FileToLoad, CFile::modeRead|CFile::shareDenyWrite, &fe) ) {
			delete pFile;
			DestroyWindow(hwnd);
			return NULL;
		}

		SendMessage(hwnd, EM_SETMARGINS, EC_LEFTMARGIN, 8);

		SendMessage(hwnd, EM_SETOLECALLBACK, 0, (LPARAM)new OleCallback);

		// richedit buffer limit -- let's set it at 16M
		ULONG lMaxSize = 0xffffff;
		SendMessage(hwnd, EM_LIMITTEXT, lMaxSize, 0);

		SetWindowLong(hwnd, GWL_USERDATA, (DWORD)pFile);

		SendMessage(hwnd, EM_SETTARGETDEVICE, 0, 0); // wrap text

		EDITSTREAM es = {0, 0, EditStreamCallBack};
		es.dwCookie = (DWORD)pFile;
		SendMessage(hwnd, EM_STREAMIN, SF_RTF, (LPARAM)&es);

		pFile->Close();

		for ( HWND hParent = hwnd; IsWindow(hParent); hParent = ::GetParent(hParent) ) {
			HMENU hMenu = GetMenu(hParent);
			if ( hMenu ) {
				hMenu = GetSubMenu(hMenu, 2);
				char buf[] = "4 Office View\t4";
				MENUITEMINFO mi;
				memset(&mi, 0, sizeof(mi));
				mi.cbSize = sizeof(mi);
				mi.fMask = MIIM_TYPE;
				mi.fType = MFT_STRING;
				mi.dwTypeData = buf;
				mi.cch = sizeof(buf);
				SetMenuItemInfo(hMenu, 3, TRUE, &mi);
			}
		}
		return hwnd;
	}
	catch (...) {
		return NULL;
	}
}

int __stdcall ListSendCommand(HWND ListWin,int Command,int Parameter)
{
	switch ( Command )
	{
	case lc_copy:
		SendMessage(ListWin, WM_COPY, 0, 0);
		return LISTPLUGIN_OK;

	case lc_newparams:
		SendMessage(ListWin, EM_SETTARGETDEVICE, 0, (Parameter & lcp_wraptext) == 0);
		return LISTPLUGIN_OK;

	case lc_selectall:
		SendMessage(ListWin, EM_SETSEL, 0 , -1);
		return LISTPLUGIN_OK;

	}
	return LISTPLUGIN_ERROR;
}

int _stdcall ListSearchText(HWND ListWin,char* SearchString,int SearchParameter)
{
	FINDTEXT find;
	int StartPos, Flags;

	if (SearchParameter & lcs_findfirst) {
		//Find first: Start at top visible line
		StartPos=SendMessage(ListWin,EM_LINEINDEX,SendMessage(ListWin,EM_GETFIRSTVISIBLELINE,0,0),0);
		SendMessage(ListWin,EM_SETSEL,StartPos,StartPos);
	} else {
		//Find next: Start at current selection+1
		SendMessage(ListWin,EM_GETSEL,(WPARAM)&StartPos,0);
		StartPos+=1;
	}

	find.chrg.cpMin=StartPos;
	find.chrg.cpMax=SendMessage(ListWin,WM_GETTEXTLENGTH,0,0);
	Flags=0;
	if (SearchParameter & lcs_wholewords)
		Flags |= FR_WHOLEWORD;
	if (SearchParameter & lcs_matchcase)
		Flags |= FR_MATCHCASE;
	if (!(SearchParameter & lcs_backwards))
		Flags |= FR_DOWN;
	find.lpstrText=SearchString;
	int index=SendMessage(ListWin, EM_FINDTEXT, Flags, (LPARAM)&find);
	if (index!=-1) {
	  int indexend=index+strlen(SearchString);
	  SendMessage(ListWin,EM_SETSEL,index,indexend);
	  int line=SendMessage(ListWin,EM_LINEFROMCHAR,index,0)-3;
	  if (line<0)
		  line=0;
      line-=SendMessage(ListWin,EM_GETFIRSTVISIBLELINE,0,0);
      SendMessage(ListWin,EM_LINESCROLL,0,line);
	  return LISTPLUGIN_OK;
	} else {
		SendMessage(ListWin,EM_SETSEL,-1,-1);  // Restart search at the beginning
	}
	return LISTPLUGIN_ERROR;
}

void __stdcall ListCloseWindow(HWND ListWin)
{
	CFile * pFile = (CFile *)GetWindowLong(ListWin, GWL_USERDATA);
	delete pFile;

	DestroyWindow(ListWin);
	return;
}

int __stdcall ListPrint(HWND ListWin,char* FileToPrint,char* DefPrinter,int PrintFlags,RECT* Margins)
{
	PRINTDLG PrintDlgRec;
	memset(&PrintDlgRec,0,sizeof(PRINTDLG));
	PrintDlgRec.lStructSize=sizeof(PRINTDLG);

    PrintDlgRec.Flags= PD_ALLPAGES | PD_USEDEVMODECOPIESANDCOLLATE | PD_RETURNDC;
	PrintDlgRec.nFromPage   = 0xFFFF; 
	PrintDlgRec.nToPage     = 0xFFFF; 
    PrintDlgRec.nMinPage	= 1;
    PrintDlgRec.nMaxPage	= 0xFFFF;
    PrintDlgRec.nCopies		= 1;
    PrintDlgRec.hwndOwner	= ListWin;// MUST be Zero, otherwise crash!
	if (PrintDlg(&PrintDlgRec)) 
	{
		HDC hdc=PrintDlgRec.hDC;
		DOCINFO DocInfo;
		POINT offset,physsize,start,avail,printable;
		int LogX,LogY;
		RECT rcsaved;

		// Warning: PD_ALLPAGES is zero!
		BOOL PrintSel=(PrintDlgRec.Flags & PD_SELECTION);
		BOOL PrintPages=(PrintDlgRec.Flags & PD_PAGENUMS);
		int PageFrom=1;
		int PageTo=0x7FFF;
		if (PrintPages) {
			PageFrom=PrintDlgRec.nFromPage;
			PageTo=PrintDlgRec.nToPage;
			if (PageTo<=0) PageTo=0x7FFF;
		}


		memset(&DocInfo,0,sizeof(DOCINFO));
		DocInfo.cbSize=sizeof(DOCINFO);
		DocInfo.lpszDocName=FileToPrint;
		if (StartDoc(hdc,&DocInfo)) {
			SetMapMode(hdc,MM_LOMETRIC);
			offset.x=GetDeviceCaps(hdc,PHYSICALOFFSETX);
			offset.y=GetDeviceCaps(hdc,PHYSICALOFFSETY);
			DPtoLP(hdc,&offset,1);
			physsize.x=GetDeviceCaps(hdc,PHYSICALWIDTH);
			physsize.y=GetDeviceCaps(hdc,PHYSICALHEIGHT);
			DPtoLP(hdc,&physsize,1);

			start.x=Margins->left-offset.x;
			start.y=-Margins->top-offset.y;
			if (start.x<0)
				start.x=0;
			if (start.y>0)
				start.y=0;
			avail.x=GetDeviceCaps(hdc,HORZRES);
			avail.y=GetDeviceCaps(hdc,VERTRES);
			DPtoLP(hdc,&avail,1);

			printable.x=min(physsize.x-(Margins->right+Margins->left),avail.x-start.x);
			printable.y=max(physsize.y+(Margins->top+Margins->bottom),avail.y-start.y);
  			
			LogX=GetDeviceCaps(hdc, LOGPIXELSX);
			LogY=GetDeviceCaps(hdc, LOGPIXELSY);

			SendMessage(ListWin, EM_FORMATRANGE, 0, 0);

			FORMATRANGE Range;
			memset(&Range,0,sizeof(FORMATRANGE));
			Range.hdc=hdc;
			Range.hdcTarget=hdc;
			LPtoDP(hdc,&start,1);
			LPtoDP(hdc,&printable,1);
			Range.rc.left = start.x * 1440 / LogX;
			Range.rc.top = start.y * 1440 / LogY;
			Range.rc.right = (start.x+printable.x) * 1440 / LogX;
			Range.rc.bottom = (start.y+printable.y) * 1440 / LogY;
			SetMapMode(hdc,MM_TEXT);

			BOOL PrintAborted=false;
			Range.rcPage = Range.rc;
			rcsaved = Range.rc;
			int CurrentPage = 1;
			int LastChar = 0;
			int LastChar2= 0;
			int MaxLen = SendMessage(ListWin,WM_GETTEXTLENGTH,0,0);
			Range.chrg.cpMax = -1;
			if (PrintPages) {
				do {
			        Range.chrg.cpMin = LastChar;
					if (CurrentPage<PageFrom) {
						LastChar = SendMessage(ListWin, EM_FORMATRANGE, 0, (LPARAM)&Range);
					} else {
						//waitform.ProgressLabel.Caption:=spage+inttostr(CurrentPage);
						//application.processmessages;
						LastChar = SendMessage(ListWin, EM_FORMATRANGE, 1, (LPARAM)&Range);
					}
					// Warning: At end of document, LastChar may be<MaxLen!!!
					if (LastChar!=-1 && LastChar < MaxLen) {
						Range.rc=rcsaved;                // Check whether another page comes
						Range.rcPage = Range.rc;
						Range.chrg.cpMin = LastChar;
						LastChar2= SendMessage(ListWin, EM_FORMATRANGE, 0, (LPARAM)&Range);
						if (LastChar<LastChar2 && LastChar < MaxLen && LastChar != -1 &&
						  CurrentPage>=PageFrom && CurrentPage<PageTo) {
							EndPage(hdc);
						}
					}

					CurrentPage++;
					Range.rc=rcsaved;
					Range.rcPage = Range.rc;
				} while (LastChar < MaxLen && LastChar != -1 && LastChar<LastChar2 &&
					     (PrintPages && CurrentPage<=PageTo) && !PrintAborted);
			} else {
				if (PrintSel) {
					SendMessage(ListWin,EM_GETSEL,(WPARAM)&LastChar,(LPARAM)&MaxLen);
					Range.chrg.cpMax = MaxLen;
				}
				do {
					Range.chrg.cpMin = LastChar;
					//waitform.ProgressLabel.Caption:=spage+inttostr(CurrentPage);
					//waitform.ProgressLabel.update;
					//application.processmessages;
					LastChar = SendMessage(ListWin, EM_FORMATRANGE, 1, (LPARAM)&Range);

					// Warning: At end of document, LastChar may be<MaxLen!!!
					if (LastChar!=-1 && LastChar < MaxLen) {
						Range.rc=rcsaved;                // Check whether another page comes
						Range.rcPage = Range.rc;
						Range.chrg.cpMin = LastChar;
						LastChar2= SendMessage(ListWin, EM_FORMATRANGE, 0, (LPARAM)&Range);
						if (LastChar<LastChar2 && LastChar < MaxLen && LastChar != -1) {
							EndPage(hdc);
						}
					}
					CurrentPage++;
					Range.rc=rcsaved;
					Range.rcPage = Range.rc;
				} while (LastChar<LastChar2 && LastChar < MaxLen && LastChar != -1 && !PrintAborted);
			}
			if (PrintAborted)
				AbortDoc(hdc);
			else
				EndDoc(hdc);
		} //StartDoc
  
		SendMessage(ListWin, EM_FORMATRANGE, 0, 0);
		DeleteDC(PrintDlgRec.hDC);
	}
	if (PrintDlgRec.hDevNames)
		GlobalFree(PrintDlgRec.hDevNames);
	return 0;
}
